DeclareRepresentation(
 "IsForestRep",
 IsComponentObjectRep and IsAttributeStoringRep,
 ["forest","alphabet","depth","nodes"]
);

InstallGlobalFunction(
 Forest,
 function(Forg)
  local F, TestForest, total, depth, maxc, Fo, Fam, alph, forestify, searchpos;
  total := 0;
  depth := 0;
  maxc := 'a';
  
  TestForest := function(L,d)
   local i;
   for i in L do
    if IsLowerAlphaChar(i) then
     total:=total+1;
     if i > maxc then maxc:=i; fi;
    elif IsList(i) and Length(i)=2 then
     if IsLowerAlphaChar(i[1]) and IsList(i[2]) then
      total:=total+1;
      if i[1] > maxc then maxc:=i[1]; fi;
      TestForest(i[2],d+1);
     else
     Error(Concatenation("Error in forest: ",String(i)," was unexpected."));
     fi;
    else
     Error(Concatenation("Error in forest: ",String(i)," was unexpected."));
    fi;
    if d>depth then depth:=d; fi;
   od;
  end;

  searchpos := function(s)
   local i, stack;
   stack := 0;
   for i in [1..Length(s)] do
    if s[i]='(' then stack:=stack+1;
    elif s[i]=')' then
     if stack=0 then return i;
     else stack:=stack-1; fi;
    fi;
   od;
   return fail; 
  end;

  forestify := function(s)
   local i, j, readNode, char, cur, sub, p, skip, ap;
   cur := []; skip:=0;
   readNode := false;

   ap := function(char, i)
    local sub;
    if IsLowerAlphaChar(char) then
     return [char, i];
    fi;
    sub := char;
    while IsBound(sub[2]) do
     sub := sub[2];
     if not IsList(sub) then Error("Syntax Error: ambiguous rooting"); fi;
     if IsList(sub[1]) then sub := sub[1]; fi;
    od;
    sub[1] := [sub[1], i];
    return char;
   end;

   for j in [1..Length(s)] do
    i := s[j];
    if skip>0 then skip:=skip-1;
    elif readNode then
     if i = '+' then
      Add(cur, char);
      readNode:=false;
     elif i='(' then
      sub := s{[j+1..Length(s)]};
      p := searchpos(sub);
      if p=fail then Error("Parse error: mismatched '(' found at position ", String(j)); fi;
      sub := sub{[1..p-1]};
      char := ap(char, forestify(sub));
      skip := p;
     elif IsLowerAlphaChar(i) then
      char := ap(char, [i]);
     else
      Error("Parse error: ", [i], " was unexpected");
     fi;
    elif IsLowerAlphaChar(i) then
     char:=i; readNode:=true;
    else
     Error("Parse error: ", [i], " unexpected at ", String(j), " in ", s);
    fi;
   od;
   if readNode then Add(cur, char); fi;
   return cur;
  end;

  if IsString(Forg) then
   F := forestify(Forg);
  else
   F := ShallowCopy(Forg);
  fi;

  TestForest(F, 0);
  alph := List([INT_CHAR('a')..INT_CHAR(maxc)], CHAR_INT);

  Fam:= NewFamily("Forest", IsForest);
  Fo := rec(
   forest := F,
   alphabet := alph,
   depth := depth,
   nodes := total
  );

  return Objectify(
   NewType(Fam, IsForest and IsForestRep and IsAttributeStoringRep),
   Fo
  );   
 end
);

InstallMethod(ViewObj,
 "displays a forest",
 true, [IsForest and IsForestRep], 0,
 function(F)
  local p,s;
  s:="";
  p := function(L)
   local i,tag;
   tag := false;
   for i in L do
    if IsLowerAlphaChar(i) then
     if tag then Add(s, '+'); else tag:=true; fi;
     Add(s, i);
    else
     if tag then Add(s, '+'); else tag:=true; fi;
     Add(s, i[1]); Add(s,'(');
     p(i[2]);
     Add(s, ')');
    fi;
   od;
  end;
  p(F!.forest);
  Print(s);
 end
);

InstallGlobalFunction(
 ForestRooting,
 function(a,F)
  if not (IsLowerAlphaChar(a) and IsForest(F)) then
   Error("Expecting a letter and forest");
  fi;

  return Forest([[a, ShallowCopy(F!.forest)]]);
 end
);

InstallGlobalFunction(
 WriteDotForest,
 function(F, file)
  local k, l, name, tdir, nome, walk, E, V, vid;

  if not IsForest(F) then
   Error("First argument must be a valid forest.");
  elif not IsString(file) then
   Error("The second argument (file name) must be a string.");
  fi;

  tdir := CMUP__getTempDir();
  name := Filename(tdir, file);
  nome := "Forest";
  E := ""; V := ""; vid := 0;

  walk := function(l, parent)
   local i;
   for i in l do
    vid := vid+1;
    if IsLowerAlphaChar(i) then
     Append(V, Concatenation(String(vid), " [shape=circle,label=",[i],"];\n"));
     if not parent=0 then
      Append(E, Concatenation(String(parent), " -> ", String(vid)," [color=black];\n"));
     else
      Append(E, Concatenation(String(vid), " -> ", String(vid)," [style=invis];\n"));
     fi;
    else
     Append(V, Concatenation(String(vid), " [shape=circle,label=",[i[1]],"];\n"));
     if not parent=0 then
      Append(E, Concatenation(String(parent), " -> ", String(vid)," [color=black];\n"));
     fi;
     walk(i[2], vid);
    fi;
   od;
  end;
  Print(E); Print(V);
  walk(F!.forest, 0);
  PrintTo(name, "digraph  ", nome, "{\n");
  AppendTo(name, E, "\n", V, "}\n");
  return tdir;
 end
);

InstallGlobalFunction(DrawForest,
 function(F)
  local gv, dot, tdir, name;

  if not IsForest(F) then
   Error("Expecting a valid forest.");
  fi;

  gv := CMUP__getPsViewer();
  dot := CMUP__getDotExecutable();
  tdir := WriteDotForest(F, "forest.dot");
  name := "forest.dot";
  CMUP__executeDotAndViewer(tdir, dot, gv, name);
 end
);

InstallMethod( \+,
 "concatenation of forests",
 [IsForest and IsForestRep, IsForest and IsForestRep],
 function(F1,F2)
  return Forest(Concatenation(F1!.forest, F2!.forest));
 end
);
