DeclareRepresentation(
 "IsForestAutomatonRep",
 IsComponentObjectRep and IsAttributeStoringRep,
 ["hfa","alphabet","output", "minimal"]
);

InstallGlobalFunction(
 ForestAutomaton,
 function(det, S, Q, A, gamma, lambda, I, F)
  local i, j, Qalph, a, FA, aut, T, L, k, l, Qm, Im,
    q, Qn, Sn, Sm, s, imgh, imgq, Tm, stacks, stackq, Lm, Fm;

  if not IsBool(det) then
   Error("\nFirst argument is a boolean (is the automaton deterministic?).\n");
  elif not IsPosInt(S) then
   Error("\nSecond argument must be a positive number of horizontal states.\n");
  elif det and not I in [1..S] then
   Error("\Seventh argument must be the number of the initial state.\n");
  elif not det and not IsList(I) then
   I := [I];
  elif not det and IsList(I) then
   for i in I do if not i in [1..S] then
    Error("\nSeventh argument must be a list of initial states.\n");
   fi; od;
  elif not IsPosInt(Q) then
   Error("\nThird argument must be a positive number of vertical states.\n");
  elif not IsPosInt(A) then
   Error("\nFourth argument must be a positive number of letters.\n");
  elif det and (not IsMatrix(gamma) or not DimensionsMat(gamma)=[Q,S]) then
   Error("\nFifth argument must be a valid transition table of the horizontal automaton.\n");
  elif det and (not IsMatrix(lambda) or not DimensionsMat(lambda)=[A,S]) then
   Error("\nSixth argument must be a valid output function.\n");
  elif not IsList(F) then
   Error("\nLast argument must be a list of final states.\n");
  fi;

  if not det then

   Qn := function(q) local i,j;
    if not IsEmpty(q) and Maximum(q)>Q then
     Error("Invalid set of vertical states in lambda: ", String(q));
    fi;
    if not q in Qm then Add(Qm, q); Add(stackq, Position(Qm,q)); fi;
    return Position(Qm, q);
   end;

   Sn := function(s)
    if not IsEmpty(s) and Maximum(s)>S then
     Error("Invalid state in horizontal DFA: ", String(s));
    fi;
    if not s in Sm then Add(Sm, s); Add(stacks, Position(Sm,s)); fi;
    return Position(Sm, s);
   end;

   imgh := function(sset, qset)
    local res, i, j;
    res := [];
    for i in sset do
     for j in qset do
      if IsBound(gamma[j][i]) then
       if IsList(gamma[j][i]) then
        Append(res, gamma[j][i]);
       else Add(res, gamma[j][i]); fi;
      fi;
     od;
    od;
    return Sn(Set(res));
   end;

   imgq := function(sset, a)
    local res, i; res:=[];
    for i in sset do
     if IsBound(lambda[a][i]) then
      if IsList(lambda[a][i]) then
       Append(res, lambda[a][i]);
      else Add(res, lambda[a][i]); fi;
     fi;
    od;
    return Qn(Set(res));
   end;

   stacks := []; stackq := [];
   Sm := []; Qm := []; Im := Sn(I);
   Fm := []; Tm := [];
   Lm := List([1..A], i->[]);

   repeat
    if not IsEmpty(stackq) then	
     for k in SortedList(stackq) do
      Tm[k] := List([1..Length(Sm)], i->imgh(Sm[i], Qm[k]));
     od;
     stackq := [];
    fi;
    if not IsEmpty(stacks) then
     k := Remove(stacks, 1);
     for a in [1..A] do
      Lm[a][k] := imgq(Sm[k], a);
     od;
     for i in [1..Length(Tm)] do
      Tm[i][k] := imgh(Sm[k], Qm[i]);
     od;
    fi;
   until IsEmpty(stacks) and IsEmpty(stackq);

   for i in [1..Length(Sm)] do
    for j in Sm[i] do
     if j in F then Add(Fm, i); break; fi;
    od;
   od;

   a := ForestAutomaton(true, Length(Sm), Length(Qm), A, Tm, Lm, Im, Fm);
   return a;

  else
   for i in lambda do
    for j in i do
     if j<1 or j>Q then Error("Output ", String(j), " is not a valid vertical state."); fi;
    od;
   od;

   if Q < 10 then
    Qalph := ""; for i in [1..Q] do Append(Qalph, String(i)); od;
   else
    Qalph := Q;
   fi;

   if A>25 then s := List([1..A], i->Concatenation("a",String(i)));
   else s := List([1..A], i -> jascii[68+i]); fi;

   a := Automaton("det", S, Qalph, gamma, [I], Set(F));
   aut := rec(
    hfa := a,
    alphabet := s,
    output := ShallowCopy(lambda),
    minimal := fail
   );
  fi;
  FA:= NewFamily("ForestAutomaton", IsForestAutomaton);
  return Objectify(
   NewType(FA, IsForestAutomaton and IsForestAutomatonRep and IsAttributeStoringRep),
   aut
  );
 end
);

InstallGlobalFunction(
 ComplementForestAutomaton,
 function(A)
  if not IsForestAutomaton(A) then Error("First argument must be a valid forest automaton"); fi;
  A!.hfa!.accepting := Difference(Set([1..A!.hfa!.states]),A!.hfa!.accepting);
  return A;
 end
);

InstallGlobalFunction(
 ForestAutomatonAccepts,
 function(A, F)
  local t,stateof;
  if not IsForestAutomaton(A) then
   Error("First argument must be a valid forest automaton");
  elif not IsForest(F) then
   Error("Second argument must be a valid forest");
  elif Length(F!.alphabet)>Length(A!.alphabet) then
   Error("Input forest uses a larger alphabet than the automaton");
  fi;
  stateof := function(A, w)
   local i, s;
   s := InitialStatesOfAutomaton(A)[1];
   for i in w do
    s := A!.transitions[i][s];
   od;
   return s;
  end;

  t := function(l)
   local i,s,j;
   s := [];
   for i in l do
    if IsList(i) then
     Add(s, A!.output[INT_CHAR(i[1])-96][t(i[2])]);
    else
     Add(s, A!.output[INT_CHAR(i)-96][1]);
    fi;
   od;
   return stateof(A!.hfa, s);
  end;
  
  return t(F!.forest) in FinalStatesOfAutomaton(A!.hfa);
 end
);

InstallGlobalFunction(
 ReachableStatesForestAutomaton,
 function(A)
  local i,j,k,l,s,q, Sm, Sn, Qm, Qn, T, L, a, bfs, I, Tm, Ign, Lm, Fm;
  if not IsForestAutomatonRep(A) then Error("Expected a forest automaton"); fi;
  
  T := A!.hfa!.transitions;
  Tm := TransposedMat(T);
  L := A!.output;
  Lm := TransposedMat(L);
  a := Length(A!.alphabet);
  I := A!.hfa!.initial[1];  
  Sm := []; Qm := []; Ign := [];
  
  bfs := function(s)
   local i,j,k, I, l, T, L;
   Add(Sm, s);
   T := Tm[s];
   L := Lm[s];
   for k in L do
    if not k in Qm then
     Add(Qm, k);
     if IsBound(Ign[k]) then
      I := Ign[k];
      Unbind(Ign[k]);
      for l in I do
       if not l in Sm then
        bfs(l);
       fi;
      od;
     fi;
    fi;
   od;
   for i in [1..Length(T)] do
    if i in Qm then
     j := T[i];
     if not j in Sm then
      bfs(j);
     fi;
    else
     if IsBound(Ign[i]) then
      Add(Ign[i], T[i]);
     else
      Ign[i] := [T[i]];
     fi;
    fi;
   od;
  end;
  
  bfs(I);
  Sort(Sm); Sort(Qm);
  Qn := i->Position(Qm, i);
  Sn := i->Position(Sm, i);

  Lm := []; Tm := [];
  for i in [1..Length(T)] do
   if i in Qm then
    l := [];
    for j in [1..Length(T[i])] do
     if j in Sm then Add(l, Sn(T[i][j])); fi;
    od;
    Add(Tm, l);
   fi;
  od;
  for i in L do
   l := [];
   for j in [1..Length(i)] do
    if j in Sm then Add(l, Qn(i[j])); fi;
   od;
   Add(Lm, l);
  od;
  
  Fm := Filtered(A!.hfa!.accepting, i->i in Sm);
  Apply(Fm, Sn);
  return ForestAutomaton(true, Length(Sm), Length(Qm), a, Tm, Lm, Sn(I), Fm);
 end
);

InstallGlobalFunction(
 MinimalForestAutomaton,
 function(Ao)
  local i, j, k, l, m, n, o, I, F, S, Q, T, L, ss, sq, subs, subq, lex,
     Tm, Lm, Sm, Qm, Fm, Im, Ti, Tij, Ns, Nq, A, Li, Tq, Ts, SQ, C, C0, C1,
     B, SS, QS, PS, PQ;

  if Ao!.minimal = fail then
  A := ReachableStatesForestAutomaton(Ao);
  S := A!.hfa!.states;
  Q := A!.hfa!.alphabet;
  I := A!.hfa!.initial[1];
  F := A!.hfa!.accepting;
  T := A!.hfa!.transitions;
  L := A!.output;
  
  Li := [];
  for i in [1..Length(A!.alphabet)] do
   Li[i] := [];
   for j in [1..S] do
    k := L[i][j];
    if IsBound(Li[i][k]) then
     if not j in Li[i][k] then Add(Li[i][k], j); fi;
    else
     Li[i][k] := [j];
    fi;
   od;
  od;

  Tq := []; Ts := [];
  for i in [1..Q] do
   Ts[i] := [];
   for j in [1..S] do
    k := T[i][j];
    if not IsBound(Tq[j]) then Tq[j]:=[]; fi;
    if IsBound(Ts[i][k]) then
     if not j in Ts[i][k] then Add(Ts[i][k], j); fi;
    else
     Ts[i][k] := [j];
    fi;
    if IsBound(Tq[j][k]) then
     Add(Tq[j][k], i);
    else
     Tq[j][k] := [i];
    fi;
   od;
  od;

  SS := []; QS := []; PS := []; PQ := [];
  if Length(F)>Int(S/2) then
   C0 := Difference([1..S], F);
   C1 := F;
  else
   C0 := F;
   C1 := Difference([1..S], F);
  fi;
  Add(SS, C1);
  PS := [C0, C1];
  PQ := [Set([1..Q])];

  while not (IsEmpty(SS) and IsEmpty(QS)) do
   if IsEmpty(QS) then
    C := Remove(SS);
    for k in [1..Length(PS)] do
     for i in Ts do
      C0 := [];
      for j in C do
       if IsBound(i[j]) then Append(C0, i[j]); fi;
      od;
      C0 := Set(C0);
      C1 := Intersection(PS[k], Difference([1..S], C0));
      C0 := Intersection(PS[k], C0);
      if IsEmpty(C0) or IsEmpty(C1) then continue; fi;
      if Length(C0)<Length(C1) then
       PS[k]:=C1; Add(PS, C0); Add(SS, C0); Add(SS,C1);
      else
       PS[k]:=C0; Add(PS, C1); Add(SS, C1); Add(SS,C0);
      fi;
     od;
    od;
    for k in [1..Length(PQ)] do
     for i in Tq do
      C0 := [];
      for j in C do
       if IsBound(i[j]) then Append(C0, i[j]); fi;
      od;
      C0 := Set(C0);
      C1 := Intersection(PQ[k], Difference([1..Q], C0));
      C0 := Intersection(PQ[k], C0);
      if IsEmpty(C0) or IsEmpty(C1) then continue; fi;
      if Length(C0)<Length(C1) then
       PQ[k]:=C1; Add(PQ, C0); Add(QS, C0);
      else
       PQ[k]:=C0; Add(PQ, C1); Add(QS, C1);
      fi;
     od;
    od;
   else
    C := Remove(QS);
    for k in [1..Length(PS)] do
     for i in Li do
      C0 := [];
      for j in C do
       if IsBound(i[j]) then Append(C0, i[j]); fi;
      od;
      C0 := Set(C0);
      C1 := Intersection(PS[k], Difference([1..S], C0));
      C0 := Intersection(PS[k], C0);
      if IsEmpty(C0) or IsEmpty(C1) then continue; fi;
      if Length(C0)<Length(C1) then
       PS[k]:=C1; Add(PS, C0); Add(SS, C0);
      else
       PS[k]:=C0; Add(PS, C1); Add(SS, C1);
      fi;
     od;
    od;
   fi;
  od;

  Sort(PS); Sort(PQ);

  subs := []; ss := [];
  for i in PS do
   for j in [2..Length(i)] do
    Add(ss, i[j]); subs[i[j]] := i[1];
   od;
  od;

  subq := []; sq := [];
  for i in PQ do
   for j in [2..Length(i)] do
    Add(sq, i[j]); subq[i[j]] := i[1];
   od;
  od;

  if IsEmpty(ss) and IsEmpty(sq) then
   A!.minimal := A;
   return A;
  fi;
  
  Sm := S-Length(ss);
  Qm := Q-Length(sq);
  Tm := [];
  Lm := [];
  Fm := Filtered(F, i->(not i in ss));
  if IsBound(subs[I]) then Im := subs[I]; else Im := I; fi;
  
  l := 0; Nq := [];
  for i in [1..Q] do
   if i in sq then continue; else l := l+1; Nq[i]:=l; fi;
  od;
  l := 0; Ns := [];
  for i in [1..S] do
   if i in ss then continue; else l := l+1; Ns[i]:=l; fi;
  od;
  Fm := List(Fm, i->Ns[i]);

  for i in [1..Q] do
   if i in sq then continue; fi;
   Ti := T[i]; Tij := [];
   for j in [1..S] do
    if not j in ss then
     if Ti[j] in ss then
      Add(Tij, Ns[subs[Ti[j]]]);
     else
      Add(Tij, Ns[Ti[j]]);
     fi;
    fi;
   od;
   Add(Tm, Tij);
  od;

  for i in L do
   Ti := [];
   for j in [1..S] do
    if not j in ss then
     if i[j] in sq then
      Add(Ti, Nq[subq[i[j]]]);
     else
      Add(Ti, Nq[i[j]]);
     fi;
    fi;
   od;
   Add(Lm, Ti);
  od;

  Ao!.minimal := ForestAutomaton(true, Sm, Qm, Length(A!.alphabet), Tm, Lm, Im, Fm);
  fi;

  return Ao!.minimal;
 end
);

InstallGlobalFunction(
 SyntacticForestAlgebra,
 function(A)
  return TransitionForestAlgebra(MinimalForestAutomaton(A));
 end
);

InstallGlobalFunction(
 TransitionForestAlgebra,
 function(A)
  local i,j,E,H,GH,n,GV,ri,li,ra, v, a1,a2, action;
  if not IsForestAutomaton(A) then
   Error("Expecting a forest automaton");
  fi;
  
  H := TransitionSemigroup(A!.hfa);
  GH := GeneratorsOfSemigroup(H);
  n := Size(H); E:=Elements(H);
  GV := [];

  for i in GH do
   ri := []; li :=[];
   for j in [1..n] do
    ra := E[j];
    ri[j] := Position(E, i*ra);
    li[j] := Position(E, ra*i);
   od;
   Add(GV, Transformation(li));
   Add(GV, Transformation(ri));
  od;

  for i in A!.output do
   ra := [];
   for j in [1..n] do
    ra[j] := Position(E, Transformation(A!.hfa!.transitions[i[1^E[j]]]));
   od;
   Add(GV, Transformation(ra));
  od;

  GV := Unique(GV);
  action := [];
  for i in GV do Add(action, ImageListOfTransformation(i)); od;

  return ForestAlgebra(H, Semigroup(GV), action);
 end
);

InstallGlobalFunction(
 HorizontalForestAutomaton,
 function(A)
  if not IsForestAutomaton(A) then
   Error("Expecting a forest automaton");
  fi;
  return ShallowCopy(A!.hfa);
 end
);

InstallMethod(ViewObj,
 "displays forest automata",
 true, [IsForestAutomaton and IsForestAutomatonRep], 0,
 function(A)
  Print(Concatenation("<Forest automaton on {",A!.alphabet,"}>"));
  ViewObj(A!.hfa); 
 end
);

InstallMethod(Display,
 "displays forest automata",
 true, [IsForestAutomaton and IsForestAutomatonRep], 0,
 function(A)
  Display(A!.hfa);
  Print("Output function :\n");
  Display(A!.output);
 end
);

InstallGlobalFunction(
 WriteDotForestAutomaton,
 function(A, fich)
  local tdir, name, T, str, out_str, footer, i, j, L, style, out, tr, a, t;

  if not IsForestAutomaton(A) or not IsString(fich) then
   Error("Invalid arguments, expecting a forest automaton and a file name.");
  fi;
  tdir := CMUP__getTempDir();
  name := Filename(tdir, Concatenation(fich, ".dot"));

  T := TransposedMat(A!.hfa!.transitions);
  L := A!.output;
  out_str := OutputTextFile(name, false);
  AppendTo(out_str, "digraph Automaton{\npad=.2;\nrankdir=LR;\n");

  for i in [1..Length(T)] do
   tr := []; a := [];
   for j in [1..Length(T[i])] do
    t := T[i][j];
    if IsBound(tr[t]) then
     tr[t] := Concatenation(tr[t], ",",String(j));
    else
     tr[t] := String(j); Add(a, t);
    fi;
   od;
   for j in a do
    AppendTo(out_str, i, " -> ", j, " [label=\"", tr[j], "\"];\n");
   od;
  od;

  for i in [1..A!.hfa!.states] do
   if i in A!.hfa!.accepting and i in A!.hfa!.initial then
    style := "Mrecord, style=filled, fillcolor=darkolivegreen3";
   elif i in A!.hfa!.accepting then
    style := "record, style=filled, fillcolor=darkolivegreen1";
   elif i in A!.hfa!.initial then
    style := "Mrecord, style=filled, fillcolor=azure1";
   else
    style := "record";
   fi;

   out := "";
   for j in [1..Length(L)] do
    if not j=1 then Append(out, "\\n"); fi;
    Append(out, Concatenation([A!.alphabet[j]]," : ", String(L[j][i])));
   od;
   AppendTo(out_str, Concatenation(String(i), " [shape=", style, ", label=\"", String(i), "|", out, "\"];\n"));
  od;

  AppendTo(out_str, "}\n");
  CloseStream(out_str);

  return tdir;
 end
);

InstallGlobalFunction(
 DrawForestAutomaton,
 function(A)
  local   gv, dot,  tdir;

  if not IsForestAutomaton(A) then
   Error("Expecting a valid forest automaton");
  fi;
    
  gv := CMUP__getPsViewer();
  dot := CMUP__getDotExecutable();
  tdir := WriteDotForestAutomaton(A, "forest_automaton");
  CMUP__executeDotAndViewer(tdir, dot, gv, "forest_automaton.dot");
 end
);

InstallGlobalFunction(
 WriteDotForestAutomatonAction,
 function(A, fich, sink)
  local tdir, name, E, n, str, H, GH, out_str, i, j, L, out, tr,
    a, t, ri, li, tag, ra, style, iscom, p, ecolor;

  ecolor := [
   "coral4","chartreuse4","cadetblue4","brown4","darkseagreen4","deepskyblue4","firebrick2","darkolivegreen2","darkturquoise","maroon","yellowgreen","mediumslateblue",
   "honeydew4","ivory2","lightcyan4","lightslategrey","lightsteelblue2","lightyellow4","lightpink3","mistyrose3","midnightblue","plum3","orchid1","orange1","thistle3",
   "wheat4", "violetred3"
  ];

  if IsInt(sink) then sink := [sink]; fi;
  if not IsForestAutomaton(A) or not IsString(fich) or not IsList(sink) then
   Error("Invalid arguments, expecting a forest automaton, a file name and sink number (0 for none) or list.");
  fi;
  tdir := CMUP__getTempDir();
  name := Filename(tdir, Concatenation(fich, ".dot"));

  H := TransitionSemigroup(A!.hfa);
  GH := GeneratorsOfSemigroup(H);
  n := Size(H); E := Elements(H);
  L := A!.output;
  
  iscom := IsCommutative(H);
  if iscom then p:=""; else p := "+"; fi;
  out_str := OutputTextFile(name, false);
  PrintTo(out_str, "digraph AutomatonAction{\nconcentrate=true;\nrankdir=LR;\n");

  for i in [1..n] do
   if i in sink then continue; fi;
   ri := []; li :=[];
   for j in [1..n] do
    ra := E[j];
    ri[j] := Position(E, E[i]*ra);
    li[j] := Position(E, ra*E[i]);
   od;

   tr := []; a := [];
   for j in [1..Length(ri)] do
    t := ri[j];
    if IsBound(tr[t]) then
     tr[t] := Concatenation(tr[t], ",",p,String(j));
    else
     tr[t] := Concatenation(p, String(j)); Add(a, t);
    fi;
   od;
   if not iscom then
    for j in [1..Length(li)] do
     t := li[j];
     if IsBound(tr[t]) then
      tr[t] := Concatenation(tr[t], ",",String(j),"+");
     else
      tr[t] := Concatenation(String(j), "+"); Add(a, t);
     fi;
    od;
   fi;

   for j in a do
    if not j in sink then
     AppendTo(out_str, i, " -> ", j, " [label=\"", tr[j], "\", fontsize=7,fontname=Helvetica];\n");
    fi;
   od;
  od;

  tr := [];
  for ri in [1..Length(L)] do
   i := L[ri]; ra := [];
   for j in [1..n] do
    if j in sink then continue; fi;
    ra[j] := Position(E, Transformation(A!.hfa!.transitions[i[1^E[j]]]));
   od;
   Add(tr, ra);
  od;

  tr := TransposedMat(tr);
  for ri in [1..Length(tr)] do
   if ri in sink then continue; fi;
   i := tr[ri]; ra := []; a := [];
   for j in [1..Length(i)] do
    t := i[j];
    if IsBound(ra[t]) then
     ra[t] := Concatenation(ra[t], ",", [A!.alphabet[j]]);
    else
     ra[t] := [A!.alphabet[j]]; Add(a, t);
    fi;
   od;
   for j in a do
    if not j in sink then
     AppendTo(out_str, ri, " -> ", j, " [label=\"", ra[j], "\"];\n");
    fi;
   od;
  od;

  for i in [1..n] do
   if not i in sink then
    if A!.hfa!.initial[1]^E[i] in A!.hfa!.accepting then
     style := "doublecircle";
    else
     style := "circle";
    fi;
    AppendTo(out_str, i, " [shape=", style,", label=\"",i,"\"];\n");
   fi;
  od;

  AppendTo(out_str, "}\n");
  CloseStream(out_str);

  return tdir;
 end
);

InstallGlobalFunction(
 DrawForestAutomatonAction,
 function(arg)
  local   A, sink, gv, dot,  tdir;

  if not IsList(arg) or not IsBound(arg[1]) or not IsForestAutomaton(arg[1]) then
   Error("Expecting a valid forest automaton");
  fi;
  A := arg[1];
  if IsBound(arg[2]) then
   sink := arg[2];
  else
   sink := 0;
  fi;

  gv := CMUP__getPsViewer();
  dot := CMUP__getDotExecutable();
  tdir := WriteDotForestAutomatonAction(A, "transition_automaton", sink);
  CMUP__executeDotAndViewer(tdir, dot, gv, "transition_automaton.dot");
 end
);

InstallGlobalFunction(
 ExistsPathInLanguageAutomaton,
 function(A, maximal)
  local i,j,l, T, L, Tr, Ad, n, a, S, I, F;

  if IsRationalExpression(A) then
   Ad := ReversedAutomaton(RatExpToAut(A));
  elif IsAutomaton(A) then
   Ad := ReversedAutomaton(A);
  else
   Error("Expecting a forest automaton or rational expression.");
  fi;
  
  Ad := UsefulAutomaton(Ad);
  Tr := ShallowCopy(Ad!.transitions);
  n := Ad!.states; I := Ad!.initial[1];
  a := Length(Tr);
  F := Ad!.accepting;
  S := [];

  if maximal then
   for i in [1..a] do
    l := List(Tr[i], j->j+1);
    Tr[i] := Concatenation([Tr[i][I]+1], l);
   od;
   n := n+1; I := 1;
   F := Set(F, i->i+1);
  fi;

  for i in [1..n] do
   if maximal then
    if i=I then
     Add(S, List([1..n], j->[j]));
    else
     Add(S, List([1..n], function(j)
      if j=I then return [i]; else return Set([i,j]); fi;
     end));
    fi;
   else
    Add(S, List([1..n], j->Set([i,j])));
   fi;
  od;

  if not maximal then
   for i in Tr do
    Apply(i, j->[j,I]);
   od;
  fi;

  return ForestAutomaton(false, n, n, a, S, Tr, I, F);
 end
);

InstallGlobalFunction(
 TreesForestAutomaton,
 function(n)
  local L, i;
  L := [];
  for i in [1..n] do Add(L, [1,1,1]); od;
  return ForestAutomaton(true, 3, 1, n, [[2,3,3]], L, 1, [2]);
 end
);

InstallGlobalFunction(
 ProductForestAutomaton,
 function(A1, A2, F)
  local L, T, i, j,a, n1, n2, q1, q2, L1, L2, p2n, q2n, T1, T2, I;
  if not (IsForestAutomatonRep(A1) and IsForestAutomatonRep(A2) and IsList(F)) then
   Error("Expecing two forest automata and a cartesian product of final states");
  fi;

  if not Length(A1!.alphabet)=Length(A2!.alphabet) then
   Error("The two automata must use the same alphabet.");
  fi;

  a := Length(A1!.alphabet);
  n1 := A1!.hfa!.states; n2 := A2!.hfa!.states;
  T1 := A1!.hfa!.transitions; T2 := A2!.hfa!.transitions;
  q1 := Length(T1); q2 := Length(T2);
  p2n := p->p[2]+(p[1]-1)*n2;
  q2n := p->p[2]+(p[1]-1)*q2;
  I := p2n([A1!.hfa!.initial[1], A2!.hfa!.initial[1]]);

  T := [];
  for i in T1 do
   for j in T2 do
    Add(T, List(Cartesian(i,j), p2n));
   od;
  od;

  L := [];
  L1 := A1!.output; L2 := A2!.output;
  for i in [1..a] do
   Add(L, List(Cartesian(L1[i],L2[i]), q2n));
  od;

  return ReachableStatesForestAutomaton(
   ForestAutomaton(true, n1*n2, q1*q2, a, T, L, I, Set(F, p2n))
  );
 end
);

InstallMethod( \*,
 "intersection of forest automata",
 [IsForestAutomaton and IsForestAutomatonRep, IsForestAutomaton and IsForestAutomatonRep],
 function(A1, A2)
  return ProductForestAutomaton(A1, A2, Cartesian(A1!.hfa!.accepting,A2!.hfa!.accepting));
 end
);

InstallMethod( \+,
 "union of forest automata",
 [IsForestAutomaton and IsForestAutomatonRep, IsForestAutomaton and IsForestAutomatonRep],
 function(A1, A2)
  local F;
  F := Concatenation(Cartesian(A1!.hfa!.accepting,[1..A2!.hfa!.states]),
    Cartesian([1..A1!.hfa!.states],A2!.hfa!.accepting));
  return ProductForestAutomaton(A1, A2, F);
 end
);

InstallMethod( AINV,
 "complement of forest automata",
 [IsForestAutomaton and IsForestAutomatonRep],
 function(A)
  return ForestAutomaton(true, A!.hfa!.states, A!.hfa!.alphabet,
   Length(A!.alphabet), A!.hfa!.transitions, A!.output, A!.hfa!.initial[1],
   Difference(Set([1..A!.hfa!.states]), A!.hfa!.accepting));
 end
);
