DeclareRepresentation(
 "IsMSORep",
 IsPositionalObjectRep and IsAttributeStoringRep,
 [2]
);

InstallGlobalFunction(
 MSOSentence,
 function(Forg)
  local F, Fo, Fam, parse, searchpos, tag, alph, forest, neg;
  alph := Set([]); forest := false;

  searchpos := function(s)
   local i, stack;
   stack := 0;
   for i in [1..Length(s)] do
    if s[i]='(' then stack:=stack+1;
    elif s[i]=')' then
     if stack=0 then return i;
     else stack:=stack-1; fi;
    fi;
   od;
   return fail; 
  end;
 
  neg := function(p)
   if p[1] in ["IMPLY","OR","AND"] then
    return [p[1], neg(p[2]), p[3]];
   else return ["NOT",,p]; fi;
  end;

  parse := function(s,start,fo,so)
   local i, j, cur, sub, p, len, ap, var;
   cur := [];
   len := Length(s);

   if len<2 then Error("Parse error: unexpected end at position ",String(start)); fi;
   if ((s[2] = '<' and s[3] = '<') or (s[2] = '<' and s[3] = '=') or (s[2] = '|' and s[3] = '|')) then
    if Length(s)>4 then s := Concatenation("(",s{[1..4]},")",s{[5..Length(s)]}); fi;
   elif (s[2] = '<' or s[2] = '.' or s[2] = '=' or s[2] = ':' or s[2] = '|') and Length(s)>3 then
    s := Concatenation("(",s{[1..3]},")",s{[4..Length(s)]});
   elif (s[1] = '^' or s[1] = '$' or s[1] = '#') and Length(s)>2 then
    s := Concatenation("(",s{[1..2]},")",s{[3..Length(s)]});
   fi;
   i := s[1];

   if i = '@' or i='&' then
    var := s[2];
    if IsUpperAlphaChar(var) then
     if var in so then Error("Parse Error: please rename variable ",[var]," at ",String(start)); fi;
     if i='@' then tag:="FORALL"; else tag:="EXISTS"; fi;
     return [tag, var, parse(s{[3..Length(s)]},start+2,fo,Concatenation(so,[var]))];
    else Error("Parse error: expecting uppercase second order variable after quantifier at ",String(start)); fi;
   elif i = '!' or i='?' then
    var :=  s[2];
    if IsLowerAlphaChar(var) then
     if var in fo then Error("Parse Error: please rename variable ",[var]," at ",String(start)); fi;
     if i='!' then tag:="forall"; else tag:="exists"; fi;
     return [tag, var, parse(s{[3..Length(s)]},start+2,Concatenation(fo,[var]),so)];
    else Error("Parse error: expecting lowercase first order variable after quantifier at ",String(start)); fi;
   elif i = '^' or i='$' then
    var :=  s[2];
    if IsLowerAlphaChar(var) then
     if not (var in fo) then Error("Parse Error: unbound variable ",[var]," at ",String(start)); fi;
     if i='^' then tag:="ROOT"; else tag:="LEAF"; fi;
     return [tag, ,var];
    else Error("Parse error: expecting lowercase first order variable after ^ or $ at ",String(start)); fi;
   elif i = '#' then
    var :=  s[2];
    if IsUpperAlphaChar(var) then
     if not (var in so) then Error("Parse Error: unbound variable ",[var]," at ",String(start)); fi;
     return ["PATH", ,var];
    else Error("Parse error: expecting uppercase second order variable after # at ",String(start)); fi;
   elif i='~' then
    sub := parse(s{[2..Length(s)]},start+1, fo, so);
    if s[2]='(' then return ["NOT", ,sub]; fi;
    return neg(sub);
   elif s[2]='.' then
    if IsLowerAlphaChar(i) and IsLowerAlphaChar(s[3]) then
     if not i in fo then Error([i]," should be a bound FO variable at ",String(start)); fi;
     if not s[3] in alph then Add(alph, s[3]); fi;
     return ["LETTER", i, s[3]];
    else Error("Parse error: syntax for . is x.a, x FO variable, a letter at",String(start)); fi;
   elif s[2]='|' then
    if (s[3] = '|') and IsLowerAlphaChar(s[4]) and IsLowerAlphaChar(i) then
     if not (i in fo and s[4] in fo) then
       Error([i]," and ",[s[4]]," should be bound FO variables at ",String(start)); fi;
     return ["ANCESTOR", i, s[4]];
    elif IsLowerAlphaChar(i) and IsLowerAlphaChar(s[3]) then
     if not (i in fo and s[3] in fo) then
       Error([i]," and ",[s[3]]," should be bound FO variables at ",String(start)); fi;
     forest := true;
     return ["PARENT", i, s[3]];
    else Error("Parse error: syntax for - is x-y, x,y FO variables at",String(start)); fi;
   elif s[2]='<' then
    if (s[3] = '<' or s[3] = '=') and IsLowerAlphaChar(s[4]) and IsLowerAlphaChar(i) then
     if not (i in fo and s[4] in fo) then
       Error([i]," and ",[s[4]]," should be bound FO variables at ",String(start)); fi;
     if s[3] = '<' then return ["PRED", i, s[4]];
     elif s[3] = '=' then return ["OR", ["EQUAL", i, s[4]], ["NEXTCHILD", i, s[4]]]; fi;
    elif IsLowerAlphaChar(i) and IsLowerAlphaChar(s[3]) then
     if not (i in fo and s[3] in fo) then
       Error([i]," and ",[s[3]]," should be bound FO variables at ",String(start)); fi;
     return ["NEXTCHILD", i, s[3]];
    else Error("Parse error: syntax for < is x<y, x,y FO variables at",String(start)); fi;
   elif s[2]='-' then
    if IsLowerAlphaChar(i) and IsLowerAlphaChar(s[3]) then
     if not (i in fo and s[3] in fo) then
       Error([i]," and ",[s[3]]," should be bound FO variables at ",String(start)); fi;
     return ["SIBLING", i, s[3]];
    else Error("Parse error: syntax for = is x=y, x,y FO variables at",String(start)); fi;
   elif s[2]='=' then
    if IsLowerAlphaChar(i) and IsLowerAlphaChar(s[3]) then
     if not (i in fo and s[3] in fo) then
       Error([i]," and ",[s[3]]," should be bound FO variables at ",String(start)); fi;
     return ["EQUAL", i, s[3]];
    else Error("Parse error: syntax for = is x=y, x,y FO variables at",String(start)); fi;
   elif s[2]=':' then
    if IsLowerAlphaChar(i) and IsUpperAlphaChar(s[3]) then
     if not (i in fo and s[3] in so) then
       Error([i]," shoud be a bound FO variable and ",[s[3]]," should be bound SO variables at ",String(start)); fi;
     return ["MEMBER", i, s[3]];
    else Error("Parse error: syntax for : is x:X, x FO and X SO variables at",String(start)); fi;
   elif i='(' then
    sub := s{[2..Length(s)]};
    p := searchpos(sub);
    if p=fail then Error("Parse error: mismatched '(' found at position ", String(start)); fi;
    sub := sub{[1..p-1]};
    s := s{[p+2..Length(s)]};
    if s="" then return parse(sub, start+1, fo, so);
    else
     if s[1] = '=' and s[2] = '>' then
      return ["IMPLY",parse(sub,start+1,fo,so),parse(s{[3..Length(s)]},start+p+4,fo,so)];
     elif s[1] = '+' then tag := "OR"; elif s[1] = '*' then tag := "AND";
     else Error("Unexpected ",[s[1]]," at position ",String(start+p+2)); fi;
     return [tag,parse(sub,start+1,fo,so),parse(s{[2..Length(s)]},start+p+3,fo,so)];
    fi;
   else
    Error("Parse error: ", [i], " was unexpected at position ",String(start));
   fi;
  end;
 
  if IsString(Forg) then
   RemoveCharacters(Forg, " ");
   F := parse(Forg, 1, [], []);
  else
   Error("Expecting a string representing a MSO sentence");
  fi;
 
  Fam:= NewFamily("MSOSentence", IsMSOSentence);
  Fo := [F, alph, forest];
 
  return Objectify(
   NewType(Fam, IsMSOSentence and IsMSORep and IsAttributeStoringRep),
   Fo
  );
 end
);

InstallMethod(ViewObj,
 "display a MSO sentence",
 true, [IsMSOSentence and IsMSORep], 0,
 function(F)
  local p;
  p := function(l)
   local i;
   i := l[1];
   if i="FORALL" then return Concatenation("@",[l[2]]," ",p(l[3]));
   elif i="EXISTS" then return Concatenation("&",[l[2]]," ",p(l[3]));
   elif i="forall" then return Concatenation("!",[l[2]]," ",p(l[3]));
   elif i="exists" then return Concatenation("?",[l[2]]," ",p(l[3]));
   elif i="NOT" then return Concatenation("~(",p(l[3]),")");
   elif i="AND" then return Concatenation("(",p(l[2]),")*(",p(l[3]),")");
   elif i="OR" then return Concatenation("(",p(l[2]),")+(",p(l[3]),")");
   elif i="IMPLY" then return Concatenation("(",p(l[2]),")=>(",p(l[3]),")");
   elif i="MEMBER" then return Concatenation([l[2]],":",[l[3]]);
   elif i="PARENT" then return Concatenation([l[2]],"|",[l[3]]);
   elif i="ANCESTOR" then return Concatenation([l[2]],"||",[l[3]]);
   elif i="NEXTCHILD" then return Concatenation([l[2]],"<",[l[3]]);
   elif i="SIBLING" then return Concatenation([l[2]],"-",[l[3]]);
   elif i="PRED" then return Concatenation([l[2]],"<<",[l[3]]);
   elif i="EQUAL" then return Concatenation([l[2]],"=",[l[3]]);
   elif i="LETTER" then return Concatenation([l[2]],".",[l[3]]);
   elif i="ROOT" then return Concatenation("^",[l[3]]);
   elif i="LEAF" then return Concatenation("$",[l[3]]);
   elif i="PATH" then return Concatenation("#",[l[3]]);
   else Error("Error in MSO sentence, unknown syntactic keyword: ",i); fi;
  end;
  Print(p(F![1]));
 end
); 

InstallGlobalFunction(
 MSOForestAutomaton,
 function(arg)
  local F, compile, alph, vwa, letter, exist, member,
   exist2, sibling, parent, equal, root, leaf, path, pred;

  if Length(arg)=0 or not (IsMSOSentence(arg[1]) and IsMSORep(arg[1])) then
   Error("Expecting a valid MSO sentence");
  fi;
  F := arg[1];
  if F![2]=[] then alph:=1; else alph := INT_CHAR(Maximum(F![2]))-96; fi;
  if Length(arg)>=2 and IsInt(arg[2]) and arg[2]>=alph then
    alph := arg[2];
  fi;

  vwa := function(fo, so)
   local V, W, A, id, T, L, i, j, sink, a;
   V := Combinations([1..Length(fo)]);
   W := Combinations([1..Length(so)]);
   A := Cartesian([1..alph],V,W);
   id := function(x,X) return Position(X,x); end;
   T := []; L := []; sink := Length(V)+1;
   for i in [1..sink] do
    T[i] := [];
    for j in [1..sink] do
     if i=sink or j=sink then T[i][j] := sink;
     elif Intersection(V[i],V[j])=[] then T[i][j] := id(Union(V[i],V[j]), V);
     else T[i][j] := sink; fi;
    od;
   od;
   for i in [1..Length(A)] do
    L[i] := [];
    for j in [1..sink] do
     a := A[i];
     if j=sink then L[i][j] := sink;
     elif Intersection(V[j],a[2])=[] then L[i][j] := id(Union(V[j],a[2]),V);
     else L[i][j]:=sink; fi;
    od;
   od;
   return MinimalForestAutomaton(
    ForestAutomaton(true, sink, sink, Length(A), T, L, id([],V), [id(List([1..Length(fo)]),V)])
   );
  end;

  letter := function(x, a, fo, so)
   local V, W, A, id, L, i, j, sink;
   V := Combinations([1..Length(fo)]);
   W := Combinations([1..Length(so)]);
   A := Cartesian([1..alph],V,W);
   L := [];
   id := function(x,X) return Position(X,x); end;
   for i in [1..Length(A)] do
    L[i] := [];
    for j in [1..2] do
     if j=2 or (A[i][1]=INT_CHAR(a)-96 and id(x,fo) in A[i][2]) then L[i][j]:=2;
     else L[i][j] := 1; fi;
    od;
   od;
   return ForestAutomaton(true, 2, 2, Length(A), [[1,2],[2,2]], L, 1, [2]);
  end;

  member := function(x, X, fo, so)
   local V, W, A, id, L, i, j;
   V := Combinations([1..Length(fo)]);
   W := Combinations([1..Length(so)]);
   A := Cartesian([1..alph],V,W);
   L := [];
   id := function(x,X) return Position(X,x); end;
   for i in [1..Length(A)] do
    L[i] := [];
    for j in [1..2] do
     if j=2 or (id(x,fo) in A[i][2] and id(X,so) in A[i][3]) then L[i][j]:=2;
     else L[i][j] := 1; fi;
    od;
   od;
   return ForestAutomaton(true, 2, 2, Length(A), [[1,2],[2,2]], L, 1, [2]);
  end;

  pred := function(x, y, fo, so)
   local V, W, A, id, L, i, j, a;
   V := Combinations([1..Length(fo)]);
   W := Combinations([1..Length(so)]);
   A := Cartesian([1..alph],V,W);
   L := [];
   id := function(x,X) return Position(X,x); end;
   for i in [1..Length(A)] do
    a := A[i];
    if id(x,fo) in a[2] then L[i] := [2, 2, 2, 4];
    elif id(y, fo) in a[2] then L[i] := [3, 4, 3, 4];
    else L[i] := [1,2,3,4]; fi;
   od;
   return ForestAutomaton(true, 4, 4, Length(A), [[1,2,3,4],[2,2,2,4],[3,4,3,4],[4,4,4,4]], L, 1, [4]);
  end;

  sibling := function(x, y, fo, so, next)
   local V, W, A, id, L, i, j, a;
   V := Combinations([1..Length(fo)]);
   W := Combinations([1..Length(so)]);
   A := Cartesian([1..alph],V,W);
   L := [];
   id := function(x,X) return Position(X,x); end;
   for i in [1..Length(A)] do
    a := A[i];
    if id(x,fo) in a[2] then j := 2;
    elif id(y,fo) in a[2] then j := 3;
    else j := 1; fi;
    L[i] := [j,j,4];
   od;
   if next then j:=1; else j:=2; fi;
   return ForestAutomaton(true, 3, 4, Length(A), [[1,j,3],[2,2,3],[1,3,3],[3,3,3]], L, 1, [3]);
  end;

  parent := function(x, y, fo, so, direct)
   local V, W, A, id, L, i, j, a;
   V := Combinations([1..Length(fo)]);
   W := Combinations([1..Length(so)]);
   A := Cartesian([1..alph],V,W);
   L := [];
   id := function(x,X) return Position(X,x); end;
   for i in [1..Length(A)] do
    a := A[i];
    if id(y,fo) in a[2] then L[i] := [2,2,3];
    elif id(x,fo) in a[2] then L[i] := [1,3,3];
    else
     if direct then j:=1; else j:=2; fi;
     L[i] := [1,j,3];
    fi;
   od;
   return ForestAutomaton(true, 3, 3, Length(A), [[1,2,3],[2,2,3],[3,3,3]], L, 1, [3]);
  end;

  root := function(x, fo, so)
   local V, W, A, id, L, i, j, a;
   V := Combinations([1..Length(fo)]);
   W := Combinations([1..Length(so)]);
   A := Cartesian([1..alph],V,W);
   L := [];
   id := function(x,X) return Position(X,x); end;
   for i in [1..Length(A)] do
    a := A[i];
    if id(x,fo) in a[2] then L[i] := [2,2];
    else L[i] := [1,1]; fi;
   od;
   return ForestAutomaton(true, 2, 2, Length(A), [[1,2],[2,2]], L, 1, [2]);
  end;

  leaf := function(x, fo, so)
   local V, W, A, id, L, i, j, a;
   V := Combinations([1..Length(fo)]);
   W := Combinations([1..Length(so)]);
   A := Cartesian([1..alph],V,W);
   L := [];
   id := function(x,X) return Position(X,x); end;
   for i in [1..Length(A)] do
    a := A[i];
    if id(x,fo) in a[2] then L[i] := [1,2,2];
    else L[i] := [1,1,2]; fi;
   od;
   return ForestAutomaton(true, 3, 2, Length(A), [[2,2,3],[3,3,3]], L, 1, [1,2]);
  end;

  path := function(X, fo, so)
   local V, W, A, id, L, i, j, a;
   V := Combinations([1..Length(fo)]);
   W := Combinations([1..Length(so)]);
   A := Cartesian([1..alph],V,W);
   L := [];
   id := function(x,X) return Position(X,x); end;
   for i in [1..Length(A)] do
    a := A[i];
    if id(X,so) in a[3] then L[i] := [2,3,2,3];
    else L[i] := [1,1,3,3]; fi;
   od;
   return ForestAutomaton(true, 4, 3, Length(A), [[2,2,3,4],[3,3,4,4],[4,4,4,4]], L, 1, [3]);
  end;

  equal := function(x, y, fo, so)
   local V, W, A, id, L, i, j, a;
   V := Combinations([1..Length(fo)]);
   W := Combinations([1..Length(so)]);
   A := Cartesian([1..alph],V,W);
   L := [];
   id := function(x,X) return Position(X,x); end;
   for i in [1..Length(A)] do
    a := A[i];
    if id(x,fo) in a[2] and id(y,fo) in a[2] then j := 2;
    else j := 1; fi;
    L[i] := [j,2];
   od;
   return ForestAutomaton(true, 2, 2, Length(A), [[1,2],[2,2]], L, 1, [2]);
  end;

  exist := function(x, As, fo, so)
   local S, Q, A, Ao, id, T, L, To, Lo, i, j, sinks, sinkq, s, q, a, F;
   S := Cartesian(List([1..As!.hfa!.states]),[0,1]);
   Q := Cartesian(List([1..As!.hfa!.alphabet]),[0,1]);
   A := Cartesian([1..alph],Combinations([1..Length(fo)]),Combinations([1..Length(so)]));
   Ao := Cartesian([1..alph],Combinations([1..Length(fo)+1]),Combinations([1..Length(so)]));

   id := function(x,X) return Position(X,x); end;
   T := []; L := []; sinks := Length(S)+1; sinkq := Length(Q)+1;
   To := As!.hfa!.transitions; Lo := As!.output;
   for i in [1..sinkq] do
    T[i] := [];
    for j in [1..sinks] do
     if i=sinkq or j=sinks then T[i][j] := sinks;
     else
      q := Q[i]; s := S[j];
      if s[2]=0 and q[2]=0 then T[i][j] := id([To[q[1]][s[1]],0], S);
      elif (q[2]=0 and s[2]=1) or (q[2]=1 and s[2]=0) then T[i][j] := id([To[q[1]][s[1]],1],S);
      else T[i][j] := sinks; fi;
     fi;
    od;
   od;
   for i in [1..Length(A)] do
    L[i] := []; a := A[i];
    for j in [1..sinks] do
     if j=sinks then L[i][j] := sinkq; else
      s := S[j];
      if s[2]=1 then L[i][j] := id([Lo[id(a,Ao)][s[1]],1],Q);
      else
       L[i][j] := [id([Lo[id(a,Ao)][s[1]],0],Q),
         id([Lo[id([a[1], Concatenation(a[2],[Length(fo)+1]), a[3]],Ao)][s[1]],1],Q)];
      fi;
     fi;
    od;
   od;
   F := List(Cartesian(As!.hfa!.accepting,[1]), function(u) return id(u,S); end);
   return ForestAutomaton(false, sinks, sinkq, Length(A), T, L, id([As!.hfa!.initial[1],0],S), F);
  end;

  exist2 := function(X, As, fo, so)
   local S, Q, A, Ao, id, L, Lo, i, j, a;
   A := Cartesian([1..alph],Combinations([1..Length(fo)]),Combinations([1..Length(so)]));
   Ao := Cartesian([1..alph],Combinations([1..Length(fo)]),Combinations([1..Length(so)+1]));
   id := function(x,X) return Position(X,x); end;

   L := []; Lo := As!.output;
   for i in [1..Length(A)] do
    L[i] := []; a := A[i];
    for j in [1..As!.hfa!.states] do
     L[i][j] := [Lo[id([a[1],a[2],Concatenation(a[3],[Length(so)+1])],Ao)][j],Lo[id(a, Ao)][j]];
    od;
   od;
   return ForestAutomaton(false, As!.hfa!.states, As!.hfa!.alphabet, Length(A), As!.hfa!.transitions,
      L, As!.hfa!.initial[1], As!.hfa!.accepting);
  end;

  compile := function(l, fo, so)
   local i, A;
   i := l[1];
   if i="FORALL" then
    return MinimalForestAutomaton(ComplementForestAutomaton(
     exist2(l[2], MinimalForestAutomaton(ComplementForestAutomaton(compile(l[3],fo,Concatenation(so,[l[2]])))*vwa(fo, Concatenation(so,[l[2]]))), fo, so)
    )*vwa(fo,so));
   elif i="EXISTS" then
    return MinimalForestAutomaton(exist2(l[2], MinimalForestAutomaton(compile(l[3],fo,Concatenation(so,[l[2]]))*vwa(fo,Concatenation(so,[l[2]]))), fo, so));
   elif i="forall" then
    return MinimalForestAutomaton(ComplementForestAutomaton(
     exist(l[2], MinimalForestAutomaton(ComplementForestAutomaton(compile(l[3],Concatenation(fo,[l[2]]),so))*vwa(Concatenation(fo,[l[2]]),so)), fo, so)
    )*vwa(fo,so));
   elif i="exists" then
    return MinimalForestAutomaton(exist(l[2], MinimalForestAutomaton(compile(l[3],Concatenation(fo,[l[2]]),so)*vwa(Concatenation(fo,[l[2]]),so)), fo, so));
   elif i="NOT" then
    return ComplementForestAutomaton(compile(l[3],fo,so));
   elif i="AND" then
    return MinimalForestAutomaton(compile(l[2],fo,so)*compile(l[3],fo,so));
   elif i="OR" then
    return MinimalForestAutomaton(compile(l[2],fo,so)+compile(l[3],fo,so));
   elif i="IMPLY" then
    return MinimalForestAutomaton(
     compile(l[3],fo,so)+ComplementForestAutomaton(compile(l[2],fo,so))
    );
   elif i="MEMBER" then
    return member(l[2],l[3],fo,so);
   elif i="PARENT" then
    return parent(l[2],l[3],fo,so, true);
   elif i="ANCESTOR" then
    return parent(l[2],l[3],fo,so, false);
   elif i="NEXTCHILD" then
    return sibling(l[2],l[3],fo,so, true);
   elif i="SIBLING" then
    return sibling(l[2],l[3],fo,so, false);
   elif i="PRED" then
    return pred(l[2],l[3],fo,so);
   elif i="ROOT" then
    return root(l[3],fo,so);
   elif i="LEAF" then
    return leaf(l[3],fo,so);
   elif i="PATH" then
    return path(l[3],fo,so);
   elif i="EQUAL" then
    return equal(l[2],l[3],fo,so);
   elif i="LETTER" then
    return letter(l[2],l[3],fo,so);
   else Error("Error in MSO sentence, unknown syntactic keyword: ",i); fi;
  end;
  return compile(F![1],[],[]);
 end
);
